/********************************************************************************/
/*																				*/
/*	host_hsb.c																	*/
/*																				*/
/*	Example host program for sending and receiving messages using HSB.			*/
/*																				*/
/*	The program sends a message containing 10 data bytes to a module in slot 1.	*/
/*	The	program then expects to receive a message from the same slot, with the	*/
/*	same data bytes sent in the message. The program then checks the received	*/
/*	data against the data that was sent.										*/
/*																				*/
/*																				*/
/*	(c) 2001 Hunt Engineering													*/
/*	    All Rights Reserved														*/
/*																				*/
/*	Version	  Date   Version Author	 Comments									*/
/*	-------	  ----   --------------	 --------									*/
/*																				*/
/*    1.1   07-12-01       JT        Changed a little to support HEPC9          */
/* 	  1.0	14-03-00	   RW		 First written.								*/
/*																				*/
/********************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <stddef.h>
#include <conio.h>
#include <time.h>
#include <dos.h>
#include <signal.h>

#include "heapi.h"

/* The following definition sets the number of data bytes in each message */
#define BUFSIZE	10


/* Function Prototypes */
int  Initialise_Host_HSB(char *dev_name, HE_WORD board);
void Error(HE_DWORD errorcode);
void Print_Win32_Error(HE_DWORD errorcode);


/* Global Variables */
HE_HANDLE   hDevice			= NULL;
HE_IOSTATUS ReadIoStatus	= NULL;
HE_IOSTATUS WriteIoStatus	= NULL;
HE_DWORD    Status;

int main(int argc, char *argv[])
{
	int i,j,k,slot;
	int status;
	int errors;
	char devname[64];
	HE_BYTE  message_type;
	HE_WORD  board;
	HE_DWORD bytes_read;
	HE_BYTE  send[BUFSIZE];
	HE_BYTE  receive[BUFSIZE];

	board = 0;
	slot  = 1;

	/* Commandline processing */
	if (argc<3)
	{
		printf("\n");
		printf("Syntax:  host_bsb device board [options]\n\n");
		printf("Options: /sx   Communicate with slot x (x=1,2,3 or 4).\n");
		printf("         The slot defaults to 1 if no /sx is used.\n\n");
		printf("E.g. host_hsb hep8a 0 /s3 tries to HSB access slot 3 on a HEPC8.\n");
		printf("E.g. host_hsb hep9a 0     tries to HSB access slot 1 on a HEPC9.\n");
		return 0;
	}

	strcpy(devname, argv[1]);
	board = atoi(argv[2]);
	for (i=3; i<argc; i++)
	{
		if (argv[i][0]=='/')
		{
			if ((argv[i][1]=='s')||(argv[i][1]=='S'))
			{
				switch (argv[i][2])
				{
				case '1' : slot = 1; break;
				case '2' : slot = 2; break;
				case '3' : slot = 3; break;
				case '4' : slot = 4; break;
				}
			}
		}
	}

	printf("Try to communicate with slot %d on %s (%d) via HSB.\n\n", slot, devname, board);

	/* Initialise the HSB interface on this board */
	status = Initialise_Host_HSB(devname, board);
	if (status < 0)
	{
		printf("Error: call to Initialise_Host_HSB failed\n");
		exit(0);
	}

	printf("At this point you must have loaded the DSP program using Code Compo-\n");
	printf("ser Studio. Then, Debug->Run (F5) or Debug->Runfree the DSP program.\n\n");

	i = 0;
	for (k=0;k<10;k++)
	{
		printf("Press return to send a message to slot %d.\n", slot);
		getchar();
		
		for (j=0;j<BUFSIZE;j++)
		{
			send[j] = i++;
			receive[j] = 0;
			if (i > 255) i = 0;
		}

		/* For this example, an arbitrary message type value of 0xaa is used */
		message_type = 0xaa;

		/* The following function call is used to send a message over HSB */
		Status = HeHSBSendMessage(hDevice, message_type, slot, send, BUFSIZE, 0);
		if (Status != 0)
		{
			printf("Error: failed to successfully send message to slot %d\n", slot);
			exit(0);
		}

		printf("Press return to receive message from slot %d\n", slot);
		getchar();

		/* The following function call is used to receive a message over HSB */
		Status = HeHSBReceiveMessage(hDevice, &message_type, slot, receive, BUFSIZE, &bytes_read, 0);
		if (Status != 0)
		{
			printf("Error: failed to successfully receive message from slot %d\n", slot);
			exit(0);
		}

		if (bytes_read != BUFSIZE)
		{
			printf("Error: expected %d data bytes, received %d\n", BUFSIZE, bytes_read);
			exit(0);
		}

		printf("Checking received data\n");
		errors = 0;
		for (j=0;j<BUFSIZE;j++)
		{
			if (send[j] != receive[j])
			{
				printf("Error: Receive[%d] = %02x\n",j,receive[j]);
				printf("\texpected %02x\n",send[i]);
				errors++;
				getchar();
			}
		}
		if (errors > 0) exit(0);
		printf("Successfully sent and received HSB messages\n");
		printf("Press return to continue\n");
		getchar();
	} 

	return(0);
}


/********************************************************************************/
/*																				*/
/*	Initialise_Host_HSB															*/
/*																				*/
/*	This function sets up the host HSB interface of each board in the system.	*/
/*																				*/
/********************************************************************************/
int Initialise_Host_HSB(char *dev_name, HE_WORD board)
{

	/* Check parameters */
	if ((board < 0) || (board > 15)) {
		printf("Error: board ID is out of range (0 to 15)\n");
		return(-1);
	}

	hDevice	  = NULL;
	ReadIoStatus  = NULL;
	WriteIoStatus = NULL;

	Status = HeOpen1(dev_name, board, "hsb\0", &hDevice);
	if (Status != HE_OK) return(1);
	
	Status = HeInitIoStatus(hDevice, &ReadIoStatus);
	if (Status != HE_OK) Error(HeGetLastOsError(hDevice));

	Status = HeInitIoStatus(hDevice, &WriteIoStatus);
	if (Status != HE_OK) Error(HeGetLastOsError(hDevice));

	return(0);
}


/********************************************************************************/
/*																				*/
/*	Error																		*/
/*																				*/
/*	General Error Routine														*/
/*																				*/
/********************************************************************************/
void Error(HE_DWORD errorcode)
{
	char text[120];


	printf("In error routine\n");
	HeErr2Text(Status, text);
	printf("Failed - %s, OsError = %ld", text, errorcode);
	Print_Win32_Error(errorcode);
	printf("\n");

	Status = HeClose(&hDevice);
	exit(1);
}


/********************************************************************************/
/*																				*/
/*	Print_Win32_Error															*/
/*																				*/
/*	Win32 Error Function														*/
/*																				*/
/********************************************************************************/
void Print_Win32_Error(HE_DWORD errorcode)
{
#ifdef _WIN32			/* This level of error code is only on WIN32 */
						/* don't even compile it if not WIN32!*/
	LPVOID lpMsgBuf;

	if(errorcode != 0) {
		printf(" (%s)\n", strerror(errorcode));

		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			errorcode, 
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), /* Default language*/
			(LPTSTR) &lpMsgBuf,
			0,
			NULL);
		printf("Sys Error - %s", lpMsgBuf);
	}
#endif				/* end of WIN32 specific error codes*/
}

