/****************************************************************************/
/* MAIN.C  - Skeleton Driver for COFF Loader       Version 6.00 4/96        */
/****************************************************************************/
/*                                                                          */
/* General:  This module is a skeleton driver for a standalone COFF         */
/*           loader.  This file is NOT part of the generic loader, but      */
/*           provides a sample interface to it that can provide a basis     */
/*           for a specific loader.  YOU MUST CUSTOMIZE THIS PROGRAM        */
/*           FOR YOUR APPLICATION.  In particular, you must write the       */
/*           function 'mem_write()' which is the low-level function to      */
/*           load target memory.                                            */
/*                                                                          */
/*           Refer to the documentation provided with the loader for        */
/*           details on how to interface with it.                           */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/* Functions defined in this module:                                        */
/*                                                                          */
/*    main()             - Main driver for loader.                          */
/*    mem_write()        - Load a buffer of raw data to the target.         */
/*    set_reloc_amount() - Determine relocation amount for each section.    */
/*    lookup_sym()       - Stub for symbol lookup routine.                  */
/*    myalloc()          - Application version of 'malloc()'.               */
/*    mralloc()          - Application version of 'realloc()'.              */
/*    load_msg()         - Called by loader to write debug messages.        */
/*                                                                          */
/****************************************************************************/

#if _LINUX
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#else
#include <windows.h>
#include <string.h>
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <time.h>
#include <dos.h>
#endif

#include "heapi.h"
#include "header.h"

#if MSDOS
#define READ_BIN "rb"
#else
#define READ_BIN "r"
#endif

/*--------------------------------------------------------------------------*/
/* LOADER DEFINITIONS                                                       */
/*--------------------------------------------------------------------------*/
FILE *fopen();
char *strrchr() /*, *malloc(), *realloc()*/;

unsigned long reloc    = 0;              /* RELOCATION AMOUNT               */
int           quiet    = 0;              /* NO BANNER                       */
int           in_read  = 0;
int           in_write = 0;

HE_HANDLE 		hDevice, hsbDevice, hReset;
HE_IOSTATUS 	ReadIoStatus, WriteIoStatus;
HE_DWORD 		Status;
HE_BOOLEAN		UseComportA;
HE_PDWORDBUFFER test;

HE_MEMHANDLE	memHandle		= NULL;

int zapsupported = 0;

#if _LINUX
void Sleep(int msec);

int stricmp(const char *s1, const char *s2)
{
   int i = 0;
   while (s1[i])
   {
      if (tolower(s1[i])!=tolower(s2[i])) return (s1[i]-s2[i]);
      i++;
   }
   return 0;
}
int GetTickCount(void)
{
   return 0;
}
#endif

void myerror(char *errstr)
{
	printf("%s", errstr);

	Status = HeClose(&hDevice);
	if(!UseComportA)
	{
		Status = HeClose(&hReset);
	}

#if _RTOS32
	printf("(Press key to continue.)\n");
	getch();
#endif

	exit(1);
}

void PrintWin32Error(HE_DWORD errorcode)
{
#ifdef _WIN32			/* This level of error code is only on WIN32  */
						/* don't even compile it if not WIN32!        */
	LPVOID lpMsgBuf;

	if (errorcode != 0)
	{
		printf(" %s)\n", strerror(errorcode));

		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			errorcode, 
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			(LPTSTR) &lpMsgBuf,
			0,
			NULL);

		printf("Sys Error - %s\n", lpMsgBuf);
	}
	else
	{
		printf("\n");
	}
#endif				/* end of WIN32 specific error codes */
}

void error(HE_DWORD errorcode)
{
	char buf[256];
	char text[120];

	HeErr2Text(Status, text);
	printf("Failed - %s, OsError = %ld", text, errorcode);
	if (Status==HE_IoInProgress)
	{
		if (in_write)
		{
			PrintWin32Error(errorcode);
			strcpy(buf, "Timeout writing board.\n");
			myerror(buf);
		}
		if (in_read)
		{
			PrintWin32Error(errorcode);
			strcpy(buf, "Timeout reading board.\n");
			myerror(buf);
		}
	}
	else
	{
		PrintWin32Error(errorcode);
		myerror(buf);
	}
}

int HeApiWrite(char *WriteBuffer, int size)
{
	HE_DWORD count;

	count    = 0;
	in_write = 1;
	Status   = HeWrite(hDevice, WriteBuffer, size, WriteIoStatus);
	if (Status == HE_IoInProgress)
	{
		while ((Status==HE_IoInProgress) && (count<25))
		{
			HeDelay(50);
			count++;
			Status = HeTestIo(hDevice, WriteIoStatus);
		}
	}
	if (Status != HE_OK)
	{
		error(HeGetLastOsError(hDevice));
		return 0;
	}
	in_write = 0;

    return 1;
}

HE_DWORD MyWriteWord(HE_HANDLE hDevice, HE_DWORD *data)
{
	HE_DWORD count;

	count    = 0;
	in_write = 1;
	Status = HeWrite(hDevice, data, 4, WriteIoStatus);
	if (Status == HE_IoInProgress)
	{
		while ((Status==HE_IoInProgress) && (count<25))
		{
			HeDelay(50);
			count++;
			Status = HeTestIo(hDevice, WriteIoStatus);
		}
	}
	if (Status != HE_OK)
	{
		error(HeGetLastOsError(hDevice));
		return 0;
	}
	in_write = 0;

	return Status;
}

HE_DWORD b0,b1,b2, BlockSize;

/*********************************************************************
 *
 *	Function : write_speed()
 *
 *********************************************************************/

void write_speed(void)
{
    HE_DWORD i, j;
#ifdef WIN32
    LARGE_INTEGER start,end, time_in, freq;
#else
    HE_DWORD start,end, time_in;
#endif

    printf("Testing...\n");

#ifdef WIN32
	QueryPerformanceFrequency(&freq);
#endif

    for (i=b0;i<=b2;i+=b1)
	{
#ifdef WIN32
		QueryPerformanceCounter(&start);
#else
		start = GetTickCount();
#endif
		if(BlockSize==0) {				/* Send a Word at a time */
			for(j=0; j<i; j++)
			{
				Status = MyWriteWord(hDevice, &test[j]);
				if(Status != HE_OK) error(HeGetLastOsError(hDevice));
			}
		}
		else {							/* Send a Block at a time */
			Status = HeWrite(hDevice, test, i<<2, WriteIoStatus);
			if(Status == HE_IoInProgress)
			{
				Status = HeWaitForIo(hDevice, WriteIoStatus);
			}
			if(Status != HE_OK) error(HeGetLastOsError(hDevice));
		}

#ifdef WIN32
		QueryPerformanceCounter(&end);
		printf("Writes Transfer size %ld DWORDS ",i);
		time_in.QuadPart = end.QuadPart - start.QuadPart;
		if (time_in.QuadPart == 0) time_in.QuadPart = 1;
		printf(   "in %d ticks, ",(int)time_in.QuadPart);
		printf(   "   Speed: %.2f KBytes/sec\n",((float)(i<<2)*(float)(freq.QuadPart))/((float)time_in.QuadPart*1024.0));
#else
		end = GetTickCount();
		printf("Writes Transfer size %ld DWORDS ",i);
		time_in = end - start;
		if (time_in == 0) time_in = 1;
		printf(   "in %d ticks, ",(int)time_in);
		printf(   "   Speed: %.2f KBytes/sec\n",((float)(i*1000.0)*4.0)/((float)time_in*1024.0));
#endif
    }

    return;
}

/* Serial Bus stuff (HSB) */

#include "hsb.h"

int hsb_error(int errorcode)
{
	char text[256];

	HeErr2Text(Status, text);
	printf("Failed - %s, OsError = %d", text, errorcode);

	if (hsbDevice!=NULL) HeClose(&hsbDevice);

	return 0;
}

int serial_bus(char *devname, int Board, int hsb_id)
{
	HE_IOSTATUS		ReadIoStatus	= NULL;
	HE_IOSTATUS		WriteIoStatus	= NULL;
	char			*Comport = "hsb";
	HE_DWORD		switches[2];
	int				count;
	HE_BYTE			msg_type;
	HE_BYTE			reply[8];
#if LOW_LEVEL
	HE_BYTE			protocol[4];
#endif

	hsbDevice		= NULL;
	ReadIoStatus	= NULL;
	WriteIoStatus	= NULL;

#if _VXWORKS
	switches[0]=HE_Switch_NoInterrupts;	
	switches[0]=HE_Switch_NoPCIMasterMode;
	switches[0]=HE_Switch_IRQ12;
	switches[1]=HE_Switch_Last;
#else
	switches[0]=HE_Switch_Last;
#endif

	Status = HeOpenS(devname, (HE_WORD)Board, Comport, &hsbDevice, switches);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	Status = HeInitIoStatus(hsbDevice, &ReadIoStatus);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	Status = HeInitIoStatus(hsbDevice, &WriteIoStatus);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	Sleep(250);

#if LOW_LEVEL
	/* Use low level HSB messages when you have non-standard or long */
	/* protocols.                                                    */
	Status = HeHSBInit(hsbDevice, 0x5);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	Status = HeHSBMaster(hsbDevice, hsb_id, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	protocol[0] = 1;			/* protocol type		*/
	protocol[1] = 0x5;			/* return HSB address	*/
	protocol[2] = 0;
	protocol[3] = 0;
	Status = HeWrite(hsbDevice, protocol, 2, WriteIoStatus);
	if(Status == HE_IoInProgress) Status = HeWaitForIo(hsbDevice, WriteIoStatus);
	if(Status != HE_OK) hsb_error(HeGetLastOsError(hsbDevice));

	Status = HeHSBSlave(hsbDevice, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	Status = HeHSBListen(hsbDevice, &i, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

    Status = HeRead(hsbDevice, reply, 7, ReadIoStatus);
    if(Status == HE_IoInProgress) Status = HeWaitForIo(hsbDevice, ReadIoStatus);
    if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	printf("Serial bus: slot %d: HERON%d-%s, rom version %d.\n", reply[1], reply[3], GetProcessor(reply[4]), reply[5]);

	Status = HeClose(&hsbDevice);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	return reply[3];
#else
	/* Use high level HSB messages for standard, short protocol queries */
	/* But standard short protocol queries can also be done with low    */
	/* level HSB functions, as shown above.                             */
	Status = HeHSBSendMessage(hsbDevice, 1, (HE_BYTE)hsb_id, NULL, 0, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

    Status = HeHSBReceiveMessage(hsbDevice, &msg_type, (HE_BYTE)hsb_id, reply, 5, (HE_DWORD *)&count, 1000);
    if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	printf("Serial bus: slot %d: HERON%d-%s, rom version %d.\n", hsb_id, reply[1], GetProcessor(reply[2]), reply[3]);

	Status = HeClose(&hsbDevice);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	return reply[1];
#endif
}

FILE *OpenOutFile(char *outpath, char *outname, char *devname, int use_bios)
{
	int   l;
	FILE *f;
	char  fname[1024];

#if _LINUX
	char  mypath[1024];

	/* Extract full path if argv[0] doesn't specify that */
	if (outpath[0]!='/')
	{
	   /* get current directory */
	   if (getcwd(fname, 1024)!=(char *)NULL)
	   {
	      strcpy(mypath, fname);
	      strcat(mypath, "/");
	      strcat(mypath, outpath);
	      strcpy(outpath, mypath);
	   }
	}
#endif
	/* We assume that outpath is a copy of argv[0]. That means that the */
	/* path includes the program name. This program name will be sliced */
	/* off in the first iteration. ------------------------------------ */
	f = NULL;
	while (f==NULL)
	{
		/* go one directory up (or down, depending on your viewpoint) */
		l = strlen(outpath);
#if _LINUX
		while ((l>=0)&&(outpath[l]!='/' )) l--;
#else
		while ((l>=0)&&(outpath[l]!='\\')) l--;
#endif
		if (l<0)
		{
			/* We can't go any deeper */
			return NULL;
		}
		outpath[l]=0;
		/* Create a new filename */
		strcpy(fname, outpath);
#if _LINUX
		strcat(fname, "/");
#else
		strcat(fname, "\\");
#endif
		if (use_bios)
		{
			if (strcmpi(devname, "hep8a")==0)
				 strcat(fname, "bios\\hepc8\\");
			else strcat(fname, "bios\\hepc9\\");
		}
		strcat(fname, outname);
		f = fopen(fname, "rb");
	}
	if (use_bios)
	{
		if (strcmpi(devname, "hep8a")==0)
			 strcat(outpath, "\\bios\\hepc8");
		else strcat(outpath, "\\bios\\hepc9");
	}
	return f;
}

int zap_heart(char *devname, HE_WORD Board)
{
	volatile int j, f;		/* volatile because of compiler problems later on,  */
	HE_BYTE      slot;
	char		 buf[4];

	Status = HeOpen(devname, Board, HSB, &hsbDevice);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	/* OK, now we have a board and we have HSB opened successfully --------------- */
	if (zapsupported)
	{
		slot   = 0x07|(Board<<3);		/* Target slot is an FPGA on board x       */
		buf[0] = 0x07;					/* secondary address: all FPGA's           */
		buf[1] = 0x07;					/* command byte                            */
		buf[2] = 22;					/* zap command                             */
		buf[3] = 0;		/* ignored */	/* What to program the timeslot reg. with  */
		Status = HeHSBStartSendMessage(hsbDevice, slot, 1000);
		if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
		Status = HeHSBSendMessageData (hsbDevice, buf, 4, 1000);
		if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
		Status = HeHSBEndOfSendMessage(hsbDevice, 1000);
		if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	}
	else
	{
		for (j=0; j<6; j++)					/* j is module id: 0, 1, 2, 3, 4, 5 ------ */
		{
			slot   = 0x07|(Board<<3);		/* Target slot is an FPGA on board x       */
			buf[1] = 0x7;					/* command byte                            */
			buf[3] = 0;						/* What to program the timeslot reg. with  */
			for (f=0; f<6; f++)				/* f is fifo 0..5                          */
			{
				int mdid = (Board<<4)|f;	/* Only used to identify the module        */
				buf[2] = f;					/* Module-to-heart fpga, timeslot register */
				buf[0] = j;					/* secondary address: Heart-to-Module FPGA */
				Status = HeHSBStartSendMessage(hsbDevice, slot, 1000);
				if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
				Status = HeHSBSendMessageData (hsbDevice, buf, 4, 1000);
				if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
				Status = HeHSBEndOfSendMessage(hsbDevice, 1000);
				if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	
				buf[0] = (0x8|j);			/* secondary address: Module-to-Heart FPGA */
				Status = HeHSBStartSendMessage(hsbDevice, slot, 1000);
				if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
				Status = HeHSBSendMessageData (hsbDevice, buf, 4, 1000);
				if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
				Status = HeHSBEndOfSendMessage(hsbDevice, 1000);
				if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
			}
		}
	}
	Status = HeClose(&hsbDevice);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	return 1;
}

int heartconfig(char *devname, HE_WORD Board, int mdid, int tslot, int hostfifo)
{
	int				host=5;					/* Module 0 is host FPGA                   */
	HE_BYTE			slot;
	int				modfifo=0;
	HE_BYTE			buf[8];

	hsbDevice		= NULL;

	Status = HeOpen(devname, Board, HSB, &hsbDevice);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	/* Program from side ------------------------------------------------------------- */
	slot   = 0x07|(Board<<3);				/* Target slot is an FPGA on board x       */
	buf[0] = 0x08|(host);					/* secondary address: Module-to-Heart FPGA */
	buf[1] = 0x7;							/* command byte                            */
	buf[2] = hostfifo;						/* Module-to-heart fpga, timeslot register */
	buf[3] = (1<<tslot);					/* What to program the timeslot reg. with  */

	Status = HeHSBStartSendMessage(hsbDevice, slot, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	Status = HeHSBSendMessageData (hsbDevice, buf, 4, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	Status = HeHSBEndOfSendMessage(hsbDevice, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	printf("Host fifo %d (out) connected to the ring, timeslot %d. [%x/%x/%x/%x]\n", hostfifo, tslot, buf[0], buf[1], buf[2], buf[3]); 

	/* Program to side --------------------------------------------------------------- */
	slot   = 0x07|(Board<<3);				/* Target slot is an FPGA on board x       */
	buf[0] = mdid;							/* secondary address: Heart-to-Module FPGA */
	buf[1] = 0x7;							/* command byte                            */
	buf[2] = modfifo;						/* Module-to-heart fpga, timeslot register */
	buf[3] = (1<<tslot);					/* What to program the timeslot reg. with  */

	Status = HeHSBStartSendMessage(hsbDevice, slot, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	Status = HeHSBSendMessageData (hsbDevice, buf, 4, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	Status = HeHSBEndOfSendMessage(hsbDevice, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	printf("Module %d fifo %d (in) connected to the ring, timeslot %d. [%x/%x/%x/%x]\n", mdid, modfifo, tslot, buf[0], buf[1], buf[2], buf[3]); 

	/* Program from side ------------------------------------------------------------- */
	slot   = 0x07|(Board<<3);				/* Target slot is an FPGA on board x       */
	buf[0] = 0x08|(mdid);					/* secondary address: Module-to-Heart FPGA */
	buf[1] = 0x7;							/* command byte                            */
	buf[2] = modfifo;						/* Module-to-heart fpga, timeslot register */
	buf[3] = (1<<(tslot));					/* What to program the timeslot reg. with  */

	Status = HeHSBStartSendMessage(hsbDevice, slot, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	Status = HeHSBSendMessageData (hsbDevice, buf, 4, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	Status = HeHSBEndOfSendMessage(hsbDevice, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	printf("Host fifo %d (in) connected to the ring, timeslot %d. [%x/%x/%x/%x]\n", hostfifo, tslot+1, buf[0], buf[1], buf[2], buf[3]); 

	/* Program to side --------------------------------------------------------------- */
	slot   = 0x07|(Board<<3);				/* Target slot is an FPGA on board x       */
	buf[0] = host;							/* secondary address: Heart-to-Module FPGA */
	buf[1] = 0x7;							/* command byte                            */
	buf[2] = hostfifo;						/* Module-to-heart fpga, timeslot register */
	buf[3] = (1<<(tslot));					/* What to program the timeslot reg. with  */

	Status = HeHSBStartSendMessage(hsbDevice, slot, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	Status = HeHSBSendMessageData (hsbDevice, buf, 4, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));
	Status = HeHSBEndOfSendMessage(hsbDevice, 1000);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	printf("Module %d fifo %d (out) connected to the ring, timeslot %d. [%x/%x/%x/%x]\n", mdid, modfifo, tslot+1, buf[0], buf[1], buf[2], buf[3]); 

	Status = HeClose(&hsbDevice);
	if(Status != HE_OK) return hsb_error(HeGetLastOsError(hsbDevice));

	return 1;
}

int pmain(int argc, char *argv[])
{
	HE_WORD		 Board;
	HE_DWORD	 config;
	HE_BOOLEAN	 ConfigSupported = FALSE;
	char		*devname;
	char		 Comport[20];
    int			 i, fifo, nul = 0, use_bios = 0;
	int			 heron_id = 1, hsb_id = 1, slot_id = 1;
	int          heron_type = -1;
	char		 outpath[1024];
	char		 outname[256];
	char		 buf[256];
	struct
	{
		int		read_ints;
		int		write_ints;
		int		mm_ints;
	} info = { 0, 0, 0 };

/*
	In the above set ConfigSupported to
  	TRUE	If you want the program to use Config (where the board supports it)
  	FALSE	If you want the program to ignore the Config signal
*/

#if _RTOS32
	/* Initialize the kernel and keyboard driver */
	RTKernelInit(5);
	KBInit();
#endif

	////////////////////////////////////////////////////////////////////////////////
	// Process the program's arguments
	////////////////////////////////////////////////////////////////////////////////
	if (argc<7)
	{
		printf("Usage: writes device board Comport Start Increment End [BlockSize] [options]\n\n");
		printf("Options:  /h1   Assume a HERON1 module.\n");
		printf("          /h4   Assume a HERON4 module.\n");
		printf("          /sx   Test the module in slot x (x=1,2,3 or 4). HEPC9 only.\n");
		printf("          /b    Use the DSP/BIOS *.out program.\n\n");
		printf("Defaults: If neither /h1 or /h4 is used, writes will try to automatically\n");
		printf("          determine the HERON type, using the HSB bus. If no slot is spe-\n");
		printf("          cified, the slot will default to 1. The default program used to\n");
		printf("          boot the DSP is a direct-access (non DSP/BIOS) program.\n\n");
		printf("E.g. writes hep8a 0 A 1000 1000 10000 /h1     for a HERON1 module on a HEPC8.\n");
		printf("E.g. writes hep9a 0 A 1000 1000 10000 /h4 /s3 for a HERON4 module in slot 3.\n");
		return 0;
	}

	devname   = argv[1];
	Board     = (HE_WORD)atoi(argv[2]);
	strcpy(Comport, "Comport");
	strcat(Comport, argv[3]);
    b0        = atol(argv[4]);
    b1        = atol(argv[5]);
    b2        = atol(argv[6]);
	BlockSize = 250;

	for (i=7; i<argc; i++)
	{
		if (argv[i][0]=='/')
		{
			if ((argv[i][1]=='h')||(argv[i][1]=='H'))
			{
				if (argv[i][2]=='1') heron_type = 1;
				if (argv[i][2]=='4') heron_type = 4;
			}
			if ((argv[i][1]=='s')||(argv[i][1]=='S'))
			{
				if (argv[i][2]=='1') slot_id    = 1;
				if (argv[i][2]=='2') slot_id    = 2;
				if (argv[i][2]=='3') slot_id    = 3;
				if (argv[i][2]=='4') slot_id    = 4;
			}
			if ((argv[i][1]=='b')||(argv[i][1]=='B'))
			{
				use_bios = 1;
			}
		}
		else
		{
			BlockSize = atol(argv[i]);
		}
	}

	if (b2 <= 0)
	{
		printf("Sorry can't end at 0! - %ld\n", b2);
		return 0;
	}
	if(b1 <= 0) b1 = b2;	// make it something sensible
	if(b0 <= 0) b0 = b1;	// make it a sensible value
	
	if(BlockSize >0)
	{
		if((b2 % BlockSize) != 0)
		{
			printf("end point (%ld) must be a multiple of block size (%ld)\n", b2, BlockSize);
			return 0;
		}
   	
		if((b1 % BlockSize) != 0)
		{
			printf("increment (%ld) must be a multiple of block size (%ld)\n", b1, BlockSize);
			return 0;
		}
   	
		if((b0 % BlockSize) != 0)
		{
			printf("start (%ld) must be a multiple of block size (%ld)\n", b1, BlockSize);
			return 0;
		}
	}
   	
	if (use_bios)
	{
		if (strcmpi(devname, "hep6a")==0)
		{
			printf("DSP/BIOS not supported for HEPC6\n");
			return 0;
		}
	}
	if (slot_id!=1)
	{
		if (stricmp(devname, "hep9a")!=0)
		{
			printf("Slot number must be 1. Only with the HEPC9 has the host access to all slots.\n");
			return 1;
		}
	}
	////////////////////////////////////////////////////////////////////////////////////////
	// End of commandline processing
	////////////////////////////////////////////////////////////////////////////////////////
   	
    UseComportA		= (tolower(argv[3][0]) == 'a');
	// The heron_id identifies a module. It is used to boot a DSP via the FIFO's.
	// The heron_id is 8 bits: bits 7/6/5/4 are the board ID (red switch), the
	// bits 3/2/1/0 are the module ID (1,2,3 or 4).
	heron_id		= ((Board&0xf)<<4)+slot_id;
	// The hsb_id identifies a module over the HSB bus. It is used to request
	// information from DSP processors over the HSB bus and also to program
	// HEART connections. The hsb_is is 7 bits: bits 6/5/4/3 are the board ID
	// (red switch) and bits 2/1/0 are the module ID (0,1,2,3,4 or 5). ID 0 is
	// to access slot 5 (the interboard connector). ID 5 is the host interface.
	hsb_id			= ((Board&0xf)<<3)+slot_id;
    
    printf("Start at %ld, inc %ld, end at %ld, BlockSize=%ld on %s (%d: %s)\n", b0, b1, b2, BlockSize, devname, Board, Comport);

	///////////////////////////////////////////////////////////////////////////////////////

	Status = HeAlloc(&memHandle, b2<<2);			/* allocate the buffer (size in BYTES) */
	if(Status != HE_OK) error(HeGetLastOsError(memHandle));

	Status = HeLock(memHandle, &test);				/* get the address */
	if(Status != HE_OK) error(HeGetLastOsError(memHandle));

	for(i=0; i<(int)b2; i++) test[i] = i;

	/*	
	To illustrate how to use the HeOpen1() routine which requires the 
	Fifo/Comport identification to be passed as a text string rather than as a 
	number we need to convert the Parameter from the command line into
	suitable text:
	*/

	/*
	The following code is an example of reset separated from
	I/O - i.e. reset on A (on an HEPC3) use Comport A or B.
	*/

	hDevice		= NULL;
	hReset		= NULL;
	
	/* ensure that we reset on A always and open it as well if we need to       */
	/* otherwise someone else could come in and assert the reset underneath us! */
	if(!UseComportA)
	{
		Status = HeOpen(devname, Board, ComportA, &hReset);
		if(Status != HE_OK) error(HeGetLastOsError(hReset));
	}

	/* open the requested Fifo or Comport */
	Status = HeOpen1(devname, Board, Comport, &hDevice);
	if(Status != HE_OK) error(HeGetLastOsError(hDevice));

    printf("Resetting...\n");
	if(UseComportA)
	{
		Status = HeReset(hDevice);
		if(Status != HE_OK) error(HeGetLastOsError(hDevice));
	}
	else
	{
		Status = HeReset(hReset);
		if(Status != HE_OK) error(HeGetLastOsError(hReset));
	}

	if (stricmp(Comport, "comporta")==0) fifo = 0;
	else
	if (stricmp(Comport, "comportb")==0) fifo = 1;
	else
	if (stricmp(Comport, "comportc")==0) fifo = 2;
	else
	if (stricmp(Comport, "comportd")==0) fifo = 3;
	else
	if (stricmp(Comport, "comporte")==0) fifo = 4;
	else
	if (stricmp(Comport, "comportf")==0) fifo = 5;
	else
	{
		printf("Error: unknown comport %s\n", Comport);
		return 0;
	}

	// for the HEPC9, we first need to configure heart:
	if (stricmp(devname,"hep9a")==0)
	{
		if (!zap_heart(devname, Board))
		{
			printf("Unable to zap HEART\n");
			return 0;
		}
		if (!heartconfig(devname, Board, hsb_id, 0 /*tslot*/, fifo))
		{
			printf("Unable to configure HEART\n");
			return 0;
		}
	}

	/* For the outfile open routine we need our copy of argv[0]      */
	strcpy(outpath, argv[0]);
	/* We also need the outfile name. This depends on whether we     */
	/* have a HERON1 or HERON4 module. Use HSB to find that out.     */
	/* Assume the out files are in the same directory as writes.exe. */
	if (stricmp(devname, "hep6a")==0)
	{
		strcpy(outname, "pc6writ.out");
	}
	else
	{
		/* If we were given a /h1 parameter, use that */
		if (heron_type==1) strcpy(outname, "writes.out");
		else
		/* If we were given a /h4 parameter, use that */
		if (heron_type==4) strcpy(outname, "writes4.out");
		else
		{
			/* No /h1 or /h4 given, use HSB to find HERON type */
			heron_type = serial_bus(devname, Board, hsb_id);
			if (heron_type==1) strcpy(outname, "writes.out");
			else
			if (heron_type==4) strcpy(outname, "writes4.out");
			else
			{
			   printf("Unknown heron type. Please specify /h1 or /h4.\n");
			   return 0;
			}
		}
	}

	ReadIoStatus	= NULL;
	WriteIoStatus	= NULL;
	
	Status = HeInitIoStatus(hDevice, &ReadIoStatus);
	if(Status != HE_OK) error(HeGetLastOsError(hDevice));

	Status = HeInitIoStatus(hDevice, &WriteIoStatus);
	if(Status != HE_OK) error(HeGetLastOsError(hDevice));

	/* ******************************************************************** */
	/*   the task of booting an application to a processor, & communicating */
	/*   with the running application starts here. It requires,		        */
	/*   1. resetting the processor                                         */
	/*   2. checking config is 0, i.e processor hasn't booted yet, optional */
	/*   3. booting code from a *.out file (a COFF file)					*/
	/*   4. checking the processor has booted, i.e config now = 1, optional */
	/*   5. communicating with the process running on the processor         */
	/*	    In this case we send blocks of data and attempt to benchm'k it  */
	/* ******************************************************************** */
	
	if(ConfigSupported)
	{
		Status = HeConfig(hDevice, &config);
		if (Status != HE_OK)
		{
			printf("Either Config not supported or some other error\n");
			ConfigSupported = FALSE;
		}
    }
    
	if(ConfigSupported)
	{
		printf("CHECKING CONFIG...\n");
		Status = HeConfig(hDevice, &config);
		if(Status != HE_OK) error(HeGetLastOsError(hDevice));
		if (config == 1)
		{
			sprintf(buf, "error. Config = 1\n");
			myerror(buf);
		}
	}

	if (!(fin = OpenOutFile(outpath, outname, devname, use_bios)))
	{
	   sprintf(buf, "\n>> can't find file <%s>\n", outname);
	   myerror(buf);
	}

   /*-----------------------------------------------------------------------*/
   /* LOAD THE FILE.  CLOSE IT WHEN FINISHED.                               */
   /*-----------------------------------------------------------------------*/
#if _LINUX
   printf("Booting %s/%s...\n", outpath, outname);
#else
   printf("Booting %s\\%s...\n", outpath, outname);
#endif

   /* For HERON boards, we must first identify what HERON to load. For the	*/
   /* HEPC6 the line below must be commented out!							*/
   if (strcmp(devname,"hep6a")!=0)
   {
	   HeApiWrite((char *) &heron_id, 4);
   }

   if (!cload())
   {
	   sprintf(buf, "\n>> error loading file\n");
	   myerror(buf);
   }
   fclose(fin);

   HeApiWrite((char *) &nul        , 4);
   HeApiWrite((char *) &entry_point, 4);

   write_speed();

   printf("Check whether any interrupts were used: ");
   Status = GetIntsStatus(hDevice, (HE_DWORD *) &info.read_ints);
   if (Status != HE_OK) error(HeGetLastOsError(hDevice));
   printf("read %x, write %x, master mode %x.\n", info.read_ints, info.write_ints, info.mm_ints);

   Status = HeUnlock(memHandle);					/* free the address */
   if(Status != HE_OK) error(HeGetLastOsError(memHandle));

   Status = HeFree(&memHandle);					/* release the buffer */
   if(Status != HE_OK) error(HeGetLastOsError(memHandle));

#if _RTOS32
   getch();
#endif

   return 1;
}


/****************************************************************************/
/*                                                                          */
/* MEM_WRITE() - Load a buffer of raw data to the target.                   */
/*                                                                          */
/*   THIS FUNCTION MUST BE CUSTOMIZED FOR YOUR APPLICATION !                */
/*                                                                          */
/*   (As supplied, this function simply dumps the data out to the screen.)  */
/*                                                                          */
/****************************************************************************/
int mem_write(buffer, nbytes, addr, page)
   unsigned char *buffer;                   /* POINTER TO DATA BUFFER       */
   unsigned int   nbytes;                   /* NUMBER OF 8-BIT BYTES        */
   T_ADDR         addr;                     /* TARGET DESTINATION ADDRESS   */
   unsigned char  page;                     /* TARGET DESTINATION PAGE      */
{
   int debug = 0;

   if (verbose)
   {
      unsigned int i;
      printf("\n[%0*lx]:%d:", 2 * sizeof(T_ADDR), addr, nbytes);
      for (i = 0; i < nbytes; ++i) printf("%02x ", buffer[i]);
      printf("\n");
   }

   /*-----------------------------------------------------------------------*/
   /* INSERT CUSTOM CODE HERE TO LOAD TARGET MEMORY.                        */
   /*-----------------------------------------------------------------------*/

   if (debug) printf("Buffer %p, size %04d, address 0x%08lx, page %04d\n", buffer, nbytes, addr, page);

   /* Round nbytes to a number that is a multiple of 4 bytes */
   /* The API will only accept sizes that are mutiples of 4  */
   nbytes = (nbytes+3)&(~3);

   HeApiWrite  (    (char *) &nbytes, 4);
   HeApiWrite  (    (char *) &addr  , 4);
   HeApiWrite  (    (char *) buffer, nbytes);

   return 1;
}


/****************************************************************************/
/*                                                                          */
/* SET_RELOC_AMOUNT() - Determine relocation amount for each section.       */
/*                                                                          */
/****************************************************************************/
int set_reloc_amount()
{
  int i;

  for (i = 0; i<  n_sections; ++i) reloc_amount[i] = reloc;
  return 1;
}


/****************************************************************************/
/*                                                                          */
/* LOOKUP_SYM() - Stub for symbol lookup routine.                           */
/*                                                                          */
/****************************************************************************/
void lookup_sym(indx, sym, aux)
   long indx;
   SYMENT *sym;
   AUXENT *aux;
{}


/****************************************************************************/
/*                                                                          */
/* LOAD_SYMS() - Stub for symbol load routine.  This routine is only called */
/*               if the global flag 'need_symbols' is TRUE.                 */
/*                                                                          */
/****************************************************************************/
int load_syms(need_reloc)
   int need_reloc;
{
   return 1;
}



/****************************************************************************/
/*                                                                          */
/* MYALLOC() - Application version of 'malloc()'.                           */
/* MRALLOC() - Application version of 'realloc()'.                          */
/*                                                                          */
/****************************************************************************/
void *myalloc(size)
   unsigned int size;
{
   char *p = malloc(size);
   if (p) return p;

   printf(   "\n>> out of memory\n");
   exit(-1);
}

void *mralloc(p, size)
   void *p;
   unsigned int size;
{
   p = realloc(p, size);
   if (p) return p;

   printf(   "\n>> out of memory\n");
   exit(-1);
}


/****************************************************************************/
/*                                                                          */
/* LOAD_MSG() - Called by loader to write debug messages.                   */
/*                                                                          */
/****************************************************************************/
void load_msg(a,b,c,d,e,g,h)
char *a;
long  b,c,d,e,g,h;
{
    printf(   a,b,c,d,e,g,h);
}
